package jumbo.xml;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.TextArea;
import java.awt.TextField;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.io.PrintWriter;
import java.io.StringReader;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import com.sun.java.swing.Icon;
import com.sun.java.swing.JButton;
import com.sun.java.swing.JComboBox;
import com.sun.java.swing.JComponent;
import com.sun.java.swing.JFrame;
import com.sun.java.swing.JLabel;
import com.sun.java.swing.JPanel;
import com.sun.java.swing.JScrollPane;
import com.sun.java.swing.JTabbedPane;
import com.sun.java.swing.JTable;

import com.sun.java.swing.text.SimpleAttributeSet;

import com.sun.java.swing.tree.DefaultMutableTreeNode;

import jumbo.xml.gui.TabbedPane;
import jumbo.xml.gui.XTable;
import jumbo.xml.gui.XText;

/** the workhorse of jumbo.xml. Most of the operations are done on XNodes or 
relate to the context of XNodes. XNodes inherit a <I>large number</I> of valuable
methods from DefaultMutableTreeNode (e.g. navigation, node structure, etc.)
, and you may wish to study this carefully.<BR>
All SAX-retained components of an XML document are XNodes (including
PCDATANodes and PINodes at present). In the future XNode will be expanded to
support the DOM (at least in part). PCDATA and PIs are subclasses of XNode).
<BR>
An XNode supports many operations common to all XML documents (e.g. navigation) and 
also provides heuristics for display (e.g. getTitle()).  An XNode has two main
parts:
<UL>
<LI>Attributes, held as a Vector (attributeVector). These can be accessed
directly, but there are also many generic operations (e.g. searching
for atttribute names and values, etc.). Attributes can be displayed in
a special component.</LI>
<LI>Content. This consists of child nodes which may be a mixture of
elements and PCDATA (along with PIs). How these are displayed depends
on a large number of heuristics which (hopefully) cover most common cases. 
These include event streams, pretty-printing, etc. XNode can be subclassed
to provide element-specific rendering.</LI>
</UL>
@author P.Murray-Rust, 1998
*/
public class XNode extends DefaultMutableTreeNode {

/**the types of displayComponent for the Node*/
// order matters as these are used elsewhere en bloc
/** the element name as a text field */
	public final static int ELEMENTNAME	= 0;
/** the attributes as a table */
	public final static int ATTRIBUTES	= ELEMENTNAME+1;
/** the CONTENT as raw XML stream */
	public final static int XMLCONTENT	= ATTRIBUTES+1;
/** heuristically rendered content; may be subclassed */
	public final static int CONTENT		= XMLCONTENT+1;
	
// the types of eventStream for the Node
/** the event stream with tags (at present identical to XML)*/
	public final static int TAGGED		= CONTENT+1;
/** the event stream with tag removed*/
	public final static int UNTAGGED	= TAGGED+1;
/** the event stream pretty printed*/
	public final static int PRETTY		= UNTAGGED+1;
/** the event stream with whitespace highlighted by comments*/
	public final static int WHITESPACE	= PRETTY+1;
	
	public final static int nComponentTypes = WHITESPACE+1;
	JComponent[] displayComponents = new JComponent[nComponentTypes];
	public final static String displayTypeStrings[] = {
		"ElementName",
		"Attributes",
		"XML Content Stream",
		"Default Style",
		"Tagged Event Stream",
		"Untagged Event Stream",
		"Pretty Printed",
		"Whitespace Highlighted",
	};
// case sensitivity and wildcard (can be OR'ed)
/** case in attributes must be matched*/
	public final static int CASE		= 0;
/** case in attributes need not be matched*/
	public final static int IGNORECASE	= 1;
/** attributes names must be matched exactly(modulated by (IGNORE)CASE)*/
	public final static int EXACT		= 0;
/** attributes names need not be matched exactly(modulated by (IGNORE)CASE)*/
	public final static int WILD		= 2;


	static Font defaultFont = new Font("Helvetica", Font.PLAIN, 12);
	static Color defaultColor = Color.blue;

	protected Vector attributeVector = null;
// the editable attribute table
	protected XTable attributeXTable = null;
// hashtable to keep track of attributes
	Hashtable attributeTable = null;
	Font font = null;
	Color color = null;
    XMLIcon xIcon = null;

	String elementName = null;
// the editable field for the element name
	protected TextField elementTextField = null;
	String title = null;

// the XText holding the content (editable)
	protected XText contentText = null;

/** highlights the elementName in content display */
	public final static SimpleAttributeSet elementStyle = 
		XText.getSimpleAttributeSet(new Font("SansSerif", Font.BOLD, 14));

/** default XNode with no name
*/
	public XNode() {
		super();
		initialise();
	}

/** XNode with  elementName defaulting to "XNode". Object is userObject
as defined in DefaultMutableTreeNode. At present I am not clear whether
I shall use it (null can be used if unsure)
@param Object object a userObject associated with the Node
*/
	public XNode(Object object) {
		this("XNode", object);
	}

/** as above, but with defined elementName
@param String elementName the elementName (as defined in xml-spec)
@param Object object a userObject associated with the Node
*/
	public XNode(String elementName, Object object) {
		super(object);
		this.elementName = elementName;
		initialise();
	}

	void initialise() {
		font = getDefaultFont();
	    color = getDefaultColor();
	}

/** copy constructor (== deep copy) XNode. This creates new child Nodes, 
attributes and content rather than simply copying references
*/
	public XNode (XNode xNode) {
		super();
//		attributeVector = new Vector();
//		attributeTable = new Hashtable();
		if (xNode.attributeVector != null) {
			for (int i = 0; i < xNode.attributeVector.size(); i++) {
				this.addAttribute(new Attribute((Attribute) xNode.attributeVector.elementAt(i)));
			}
		}
		copyMemberData(xNode);
		for (int i = 0; i < xNode.getChildCount(); i++) {
			Object child = xNode.getChildAt(i);
			if (child instanceof PINode) {
				this.add(new PINode((PINode) child));
			} else if (child instanceof PCDATANode) {
				this.add(new PCDATANode((PCDATANode) child));
			} else if (child instanceof XNode) {
				this.add(new XNode((XNode) child));
			}
		}
	}

	void copyMemberData(XNode fromNode) {
		font = fromNode.font;
		color = fromNode.color;
		xIcon = fromNode.xIcon;
	    elementName = fromNode.elementName;
	    title = fromNode.title;
	}

/** replace this XNode with another; if this==root, new tree is created.
If there is an associated tree, it will be reset with resetTree()
@param XNode newNode node to replace this.XNode in tree
*/
	public void replaceWith(XNode newNode) {
		XTree xTree = XTree.getTree(this);
		XNode parent = (XNode) this.getParent();
		if (parent != null) {
			int iChild = parent.getIndex(this);
			parent.insert(newNode, iChild);
			parent.remove(this);
		}
		if (xTree != null) {
			xTree.resetTree(newNode);
		} else {
			System.out.println("Cannot reset null tree");
		}
	}

/** add an attribute; if name already exists, overwrites it
@param String name the attribute name (as defined in xml-spec)
@param String value the attribute value (as defined in xml-spec)
@param int type the attribute type (as defined in xml-spec). Possible values are
CDATA, ID, IDREF, IDREFS, NMTOKEN, NMTOKENS, NOTATION and NDATA
@return boolean true if attribute already exists
*/
	public boolean addAttribute(String name, String value, int type) {
		Attribute attribute = new Attribute(name, value, type);
		return this.addAttribute(attribute);
	}

/** add an attribute; if name already exists, overwrites it. If attribute
is null, no-op.
@param Attribute the attribute 
@return boolean true if attribute already exists (and therefore has been changed)
*/
	public boolean addAttribute(Attribute attribute) {
		if (attribute == null) return false;
		boolean exists = false;
		if (attributeVector == null) {
			attributeTable = new Hashtable();
			attributeVector = new Vector();
		}
		Attribute oldAttribute = (Attribute) attributeTable.get(attribute.name);
		if (oldAttribute != null) {
			this.removeAttribute(attribute.name);
			exists = true;
		}
		attributeVector.addElement(attribute);
		attributeTable.put(attribute.name, attribute);
		return exists;
	}

/** delete an attribute. If name is null, or not found, no action.
@param String name the attribute name (as defined in xml-spec)
*/
	public void removeAttribute(String name) {
		if (name == null || attributeVector == null) return;
		Attribute oldAttribute = (Attribute) attributeTable.get(name);
		if (oldAttribute != null) {
			attributeTable.remove(name);
			int sz = attributeVector.size();
			for (int i = 0; i < sz; i++) {
				Attribute attribute = (Attribute) attributeVector.elementAt(i);
				if (attribute.name.equals(name)) {
					attributeVector.removeElementAt(i);
					return;
				}
				
			}
		}
	}

/** returns a Vector containing the attributes. If none, returns null
@return Vector attributeVector (contains zero or more Attributes)
*/
	public Vector getAttributeVector() {return attributeVector;}
	
/** returns the attributes in concatenated text form (suitable for
insertion to start tags). If none, returns "". Example:<BR>
xml:link="simple" href="http://www.some/where"<BR>
@return String the attribute String
*/
	public String getAttributeString() {
		String s = "";
		if (attributeVector != null) {
			int sz = attributeVector.size();
			for (int i = 0; i < sz; i++) {
				Attribute att = (Attribute) attributeVector.elementAt(i);
				s += " "+att.toString();
			}
		}
		return s;
	}

/** gets the attribute with a given name (ignored if name is
null or bad value of sensitivity)
@param String name the attribute name
@param int sensitivity can take values CASE or IGNORECASE
@return Attribute the attribute (or null if not found)
*/
	public Attribute getAttribute(String name, int sensitivity) {
		if (name == null || attributeVector == null) return null;
		if (sensitivity == CASE) {
			return (Attribute) attributeTable.get(name);
		}
		int nAttributes = attributeVector.size();
		for (int i = 0; i < nAttributes; i++) {
			Attribute attribute = (Attribute) attributeVector.elementAt(i);
			if (attribute.name.equalsIgnoreCase(name)) return attribute;
		}
		return null;
	}
			
/** gets the value for an attribute with a given name
(ignored if name is null or bad value of sensitivity)
@param String name the attribute name
@param int sensitivity can take values CASE or IGNORECASE
@return Attribute the attribute (or null if not found)
*/
	public String getAttributeValue(String name, int sensitivity) {
		Attribute attribute = getAttribute(name, sensitivity);
		if (attribute == null) return null;
		return attribute.value;
	}
			
/** gets first child with given element name
(ignored if name is null or bad value of sensitivity). In principle
you could search for PCDATA or PI but we shall provide other methods
@param String name the elementName (must match exactly other than case)
@param int sensitivity can take values CASE or IGNORECASE
@return XNode the child 
*/
	public XNode getChildWithElementName(String name, int sensitivity) {
		if (name == null || getChildCount() == 0) return null;
		Enumeration childs = children();
		while (childs.hasMoreElements()) {
			XNode child = (XNode) childs.nextElement();
			if ((sensitivity == CASE && child.elementName.equals(name)) ||
			    (sensitivity == IGNORECASE && child.elementName.equalsIgnoreCase(name))) {
				return child;
			}
		}
		return null;
	}

/** gets first child with given value of attribute
(ignored if name is null or bad value of sensitivity). 
@param String name the attributeName (must match exactly other than case)
@param int nameSensitivity can take values CASE or IGNORECASE
@param String value the attributeValue (must match exactly other than case)
@param int valueSensitivity can take values CASE or IGNORECASE
@return XNode the child 
*/
	public XNode getChildWithAttributeValue
		(String name, int nameSensitivity, String value, int valueSensitivity) {
		if (name == null || getChildCount() == 0) return null;
		Enumeration childs = children();
		while (childs.hasMoreElements()) {
			XNode child = (XNode) childs.nextElement();
			String attributeValue = child.getAttributeValue(name, nameSensitivity);
			if (attributeValue != null) {
			    if (valueSensitivity == CASE && attributeValue.equals(value)) {
					return child;
				}
			    if (valueSensitivity == IGNORECASE && attributeValue.equalsIgnoreCase(value)) {
					return child;
				}
			}
		}
		return null;
	}

/** replace a child XNode by its specialised version. Thus if the child
has an elementType of FOO a new FOONode is made to which are transferred
all the attributes and content of child. In this way, specialised Nodes
can be made after the tree is constructed. Experimental.
@param XNode xNode node to be transformed. essentially destroyed
@param String subClassName class of specialied Node
@return XNode new Node of class subClassName (null if error)
@exception XException various (subClass is not subClass of XNode; bugs)
*/
	public static XNode makeSubClassedNode(XNode xNode, String subClassName) 
		throws XException {
		if (xNode == null) return null;
		XNode newNode = null;
		try {
			Class subClass = Class.forName(subClassName);
			newNode = (XNode) subClass.newInstance();
		} catch (ClassCastException e) {
			throw new XException(e);
		} catch (IllegalAccessException e1) {
			throw new XException(e1);
		} catch (InstantiationException e2) {
			throw new XException(e2);
		} catch (ClassNotFoundException e3) {
			throw new XException(e3);
		}
		newNode.attributeVector = xNode.attributeVector;
		xNode.attributeVector = null;
		newNode.attributeTable = xNode.attributeTable;
		xNode.attributeTable = null;
		newNode.displayComponents = xNode.displayComponents;
		xNode.displayComponents = null;
		newNode.copyMemberData(xNode);
// transfer children
		int nChild = xNode.getChildCount();
		for (int i = 0; i < nChild; i++) {
			try {
				newNode.add((XNode)xNode.getChildAt(0));
			} catch (ClassCastException e4) {
				throw new XException(e4);
			}
		}
// attach new node to parent and delete old one
		xNode.replaceWith(newNode);
		return newNode;
	}

/** heuristics for getting title
*/
	static String[] titleAttributes = {
		"TITLE",
		"CHAPTER",
		"SECTION",
		"CLASS",
		"ID",
	};

/** gets a title for the XNode. There are several cascading heuristics here:
<UL>
<LI>If this is a PCDATA node, return the content value</LI>
<LI>If the PCDATA is whitespace, returns "&lt;!----&gt;"</LI>
<LI>If this has an attribute with a common name (e.g. TITLE, ID, etc.)
uses the attribute value. These names may be resettable in later versions.
</LI>
<LI>If this has a child with a common elementName (e.g. TITLE) and the
child contains only a PCDATA node, returns the string content</LI>
<LI>Return elementName if not null</LI>
<LI>Return the (Java) class of the node</LI>
</UL>
This will always result in a non-null title (i.e. "XNode"). The title
is cached.
@return String the title
*/

	public String getTitle() {
		if (title != null) return title;
		if (this instanceof PCDATANode) {
			title = ((PCDATANode)this).content;
			if (title.trim().equals("")) {
				title = "<!---->";
			} else {
				title = "[S]"+title;
			}
		}
// search for common attribute types
		if (title == null) {
			for (int i = 0; i < titleAttributes.length; i++) {
				title = getAttributeValue(titleAttributes[i], IGNORECASE);
				if (title != null) {
					break;
				}
			}
		}
// search for child with TITLE elementName
		if (title == null) {
			XNode child = getChildWithElementName("TITLE", IGNORECASE);
			if (child != null) title = child.getPCDATAContent();
		}		
		if (title == null) {
			title = elementName;
			String s = getPCDATAContent();
			if (s != null) title += ": "+s;
		}
		if (title == null) title = getClass().getName();
		return title.substring(0, Math.min(30, title.length()));
	}

/** if this has a single PCDATA child, returns its string value, else null
@return String String content of single PCDATA child
*/
	public String getPCDATAContent() {
		if (getChildCount() == 1) {
			XNode child = (XNode) getChildAt(0);
			if (child instanceof PCDATANode) {
				return ((PCDATANode)child).content;
			}
		}
		return null;
	}

/** is the XNode an element Node (as opposed to a PCDATA node or
PI or other...
*/
	public boolean isElementNode() {
		return !(elementName.startsWith("#"));
	}

/** return the elementName
@return String element name
*/
	public String getElementName() {
		return elementName;
	}

/** gets the event stream (as XML or text). Options for <TT>type</TT><BR>
<UL>
<LI>XMLCONTENT rendering of content as XML</LI>
<LI>TAGGED	precise rendering of XML (at present == XML)</LI>
<LI>UNTAGGED all tags removed</LI>
<LI>PRETTY pretty printed (indented tags)</LI>
<LI>WHITESPACE with whitespace elements highlighted by comments</LI>
</UL>
@param int type	type chosen from list above
@param int level of indentation (normally set to 0 when called, and expanded
recursively internally);
@return String the event stream, escaped where needed. If there are newlines, etc. they are 
embedded in the stream.  
*/
	public String getEventStream(int type, int level) {
		if (this instanceof PINode) {
			if (type != UNTAGGED) {
				PINode piNode = (PINode) this;
				return "<?"+piNode.name+" "+piNode.remainder+"?>";
			}
		} else if (this instanceof PCDATANode) {
			String s = ((PCDATANode) this).content;
			s = Util.escape(s);
			if (type == PRETTY) {
				s = s.trim();
			} else if (type == WHITESPACE) {
				if (s.trim().equals("")) {
					s = "<!---->"+s;
				}
			}
			return s;
		}
		StringBuffer sb = new StringBuffer();
// if only a single PCDATANode, omit returns in prettyprinting
		boolean prettyCR = true;
		if (type == PRETTY && getPCDATAContent() != null)	{
			prettyCR = false;
		}
		int nspace = level*2;
		if (type == PRETTY) {
			sb.append(Util.spaces(nspace));
		}
		if (type == UNTAGGED || (type == XMLCONTENT && level == 0)) {
		} else {
			sb.append("<"+elementName+getAttributeString()+">");
		}
		if (type == PRETTY && prettyCR) {
			sb.append("\n");
		}
		//level++;
		int count = 0;
		Enumeration childs = this.children();
		while (childs.hasMoreElements()) {
			Object child = childs.nextElement();
			if (false) {
			} else if (child instanceof XNode) {
				String ss = ((XNode)child).getEventStream(type, level+1);
				sb.append(ss);
			}
		}
		if (type == PRETTY && prettyCR) {
			sb.append(Util.spaces(nspace));
		}
		if (type == UNTAGGED || (type == XMLCONTENT && level == 0)) {
		} else {
			sb.append("</"+elementName+">");
		}
		if (type == PRETTY) {
			sb.append("\n");
		}
		return sb.toString();
	}
/**  as for getEventStream, but writes to a PrintWriter. It must be
possible to avoid the duplication...
*/

	public void writeEventStream(PrintWriter pw, int type, int level) {
		if (this instanceof PINode) {
			if (type != UNTAGGED) {
				PINode piNode = (PINode) this;
				pw.println("<?"+piNode.name+" "+piNode.remainder+"?>");
				return;
			}
		} else if (this instanceof PCDATANode) {
			String s = ((PCDATANode) this).content;
			s = Util.escape(s);
			if (type == PRETTY) {
				s = s.trim();
			} else if (type == WHITESPACE) {
				if (s.trim().equals("")) {
					s = "<!---->"+s;
				}
			}
			pw.print(s);
			return;
		}
//		StringBuffer sb = new StringBuffer();
// if only a single PCDATANode, omit returns in prettyprinting
		boolean prettyCR = true;
		if (type == PRETTY && getPCDATAContent() != null)	{
			prettyCR = false;
		}
		int nspace = level*2;
		if (type == PRETTY) {
//			sb.append(Util.spaces(nspace));
			pw.print(Util.spaces(nspace));
		}
		if (type != UNTAGGED) {
//			sb.append("<"+elementName+getAttributeString()+">");
			pw.print("<"+elementName+getAttributeString()+">");
		}
		if (type == PRETTY && prettyCR) {
//			sb.append("\n");
			pw.print("\n");
		}
		level++;
		int count = 0;
		Enumeration childs = this.children();
		while (childs.hasMoreElements()) {
			Object child = childs.nextElement();
			if (false) {
			} else if (child instanceof XNode) {
//				String ss = ((XNode)child).getEventStream(type, level);
				((XNode)child).writeEventStream(pw, type, level);
//				sb.append(ss);
			}
		}
		if (type == PRETTY && prettyCR) {
//			sb.append(Util.spaces(nspace));
			pw.print(Util.spaces(nspace));
		}
		if (type != UNTAGGED) {
//			sb.append("</"+elementName+">");
			pw.print("</"+elementName+">");
		}
		if (type == PRETTY) {
//			sb.append("\n");
			pw.print("\n");
		}
//		return sb.toString();
		return;
	}

/** dummy, to be overridden
*/
	public void display() {
	}
	
/** dummy, to be overridden by specialised classes
*/
	public void onClick() {
	}
	

/** returns a choice of document components, for different 'views'
of the node. Values are ATTRIBUTE, XMLCONTENT, TAGGED, etc. The returned
component is a fully functioning (e.g. scrollable) component 
that can be add()ed to a Container.
@param int type type of component to return
@return JComponent the component 
*/
	public JComponent getDisplayComponent(int type) {
		if (displayComponents[type] == null) {
			XText xText = null;
			JComponent component = null;
			try {
				if (type == CONTENT) {
					xText = new XText();
					addContent(xText, 0);
					xText.setEditable(false);
				} else if (type == ELEMENTNAME) {
					component = new JPanel();
					component.setLayout(new BorderLayout());
					component.add(new JLabel("ElementName: "), BorderLayout.WEST);
					elementTextField = new TextField(elementName);
					component.add(elementTextField, BorderLayout.CENTER);
					JButton button = new JButton("Update");
					button.setEnabled(true);
					ActionListener al = new ActionListener() {
						public void actionPerformed(ActionEvent e) {
							updateTree();
						}
					};
					button.addActionListener(al);
					component.add(button, BorderLayout.EAST);
				} else if (type == ATTRIBUTES) {
					JPanel jPanel = new JPanel();
					jPanel.setLayout(new BorderLayout());
					JPanel topPanel = new JPanel();
					topPanel.setLayout(new BorderLayout());
					jPanel.add(topPanel, BorderLayout.NORTH);

					if (attributeVector == null) {
						attributeVector = new Vector();
						Attribute dummyAttribute = new Attribute("", "", Attribute.CDATA);
						attributeVector.addElement(dummyAttribute);
					}
					String columnNames[] = {"Name", "Value", "Type"};
					int natts = attributeVector.size();
					Object data[][] = new Object[natts][];
					for (int iRow = 0; iRow < natts; iRow++) {
						data[iRow] = new Object[3];
						Attribute att = (Attribute) attributeVector.elementAt(iRow);
						data[iRow][0] = att.name;
						data[iRow][1] = att.value;
						data[iRow][2] = Attribute.typeString[att.type];
					}
					try {
						attributeXTable = new XTable("Attributes", data, columnNames);
						attributeXTable.addComboBoxEditor("Type", Attribute.typeString);
						Component tableComponent = attributeXTable.getDisplayComponent();
						jPanel.add(tableComponent, BorderLayout.CENTER);
						component = jPanel;
					} catch (Exception e) {Util.bug(e);}
// delete attribute
					JButton jButton = new JButton("Delete Attribute");
					topPanel.add(jButton, BorderLayout.EAST);
					final XTable x1Table = attributeXTable;
			        ActionListener al = new ActionListener() {
				        public void actionPerformed(ActionEvent e) {
					        int selectedRow = x1Table.getSelectedRow();
					        if (selectedRow == -1) {
						        System.out.println("No row selected");
					        } else {
						        x1Table.removeRow(selectedRow);
					        }
				        }
			        };
					jButton.addActionListener(al);
// add attribute
					jButton = new JButton("Add Attribute");
					topPanel.add(jButton, BorderLayout.WEST);
					al = new ActionListener() {
						public void actionPerformed(ActionEvent e) {
							Object[] row = {"", "", Attribute.typeString[Attribute.CDATA]};
							x1Table.addRow(row);
						}
					};
					jButton.addActionListener(al);
				} else if (type == XMLCONTENT ||
				    type == PRETTY ||
				    type == TAGGED ||
				    type == UNTAGGED ||
				    type == WHITESPACE) {
					xText = new XText();
					xText.addText(this.getEventStream(type, 0), new SimpleAttributeSet());
					xText.setEditable(false);
					if (type == XMLCONTENT) xText.setEditable(true);
				}
				if (xText != null) {
					displayComponents[type] = new JScrollPane(xText.getJTextPane());
					if (type == XMLCONTENT) contentText = xText;
				} else if (component != null) {
					displayComponents[type] = component;
				}
			} catch (Exception e) {e.printStackTrace();}
		}
		return displayComponents[type];
	}
		
    private void updateTree() {
// build up XML
		String xmlStream = "<";
		if (elementTextField == null) {
			System.out.println("*** PROBLEM NO Element Name ***");
			return;
		}
		String elementName = elementTextField.getText();
		xmlStream += elementName;
		if (attributeXTable != null) {
			for (int i = 0; i < attributeXTable.getRowCount(); i++) {
				Vector row = attributeXTable.getRow(i);
				if (row.size() < 2) {
					System.out.println("BAD ATTRIBUTE COLUMN COUNT");
					return;
				}
// only need name and value; ignore blank attname
				String attName = (String) row.elementAt(0);
				if (attName.trim().equals("")) continue;
				xmlStream += " "+attName+"=\"";
				xmlStream += Util.escape((String)row.elementAt(1))+"\"";
			}
		}
		xmlStream += ">";
		String content = "";
		if (contentText != null) {
			content = contentText.getText();
		} else {
			content = this.getEventStream(XNode.XMLCONTENT, 0);
		}
		xmlStream += content+"</"+elementName+">";
		SAX sax = null;
		try {
		    sax = new SAX(new StringReader(xmlStream));
		} catch (Exception e) {
			System.out.println("SAX PARSE FAILURE: "+e);
			return;
		}											
		XTree newTree = sax.getTree();
// replace current Node
		XNode newNode = newTree.getRootNode();
		this.replaceWith(newNode);
		newNode.title = null;
	}

/** add component display to an existing Container. Will create a
JComponent through getDisplayComponent() above. If Container is a
TabbedPane, will create toolTips according to the type.
@param Container container the container to add the component to
@param int the component type (XMLCONTENT, TAGGED, etc.) 
*/
	public void addDisplayComponentTo(Container container, int type) {
		Component displayComponent = this.getDisplayComponent(type);
		if (displayComponent == null) return;
// adding to a TabbedPane?
		if (container instanceof TabbedPane) {
			String title = this.getTitle();
			String toolTip = "";
			if (type == CONTENT) {
				toolTip = "C";
				Object[] objs = this.getPath();
				for (int i = 0; i < objs.length; i++) {
					toolTip += "/"+objs[i].toString();
				}
			} else if (type == ATTRIBUTES) {
				toolTip = "Attributes: "+title;
			} else if (type == XMLCONTENT) {
				toolTip = "XML Content: "+title;
			}
			((TabbedPane) container).add(displayComponent, title, toolTip);  
		} else {
			container.removeAll();
			container.add(displayComponent);
		}
	}

/** default formatting of content in absence of stylesheet. Often
likely to be overridden. 
@param XText xText an existing XText JComponent. The content will be
appended to the text already inside
@param int level the level of recursion (use 0 by default)
*/
	protected void addContent(XText xText, int level) {
		if (this instanceof PCDATANode) {
			xText.addText(((PCDATANode)this).content, XText.defaultStyle);
			return;
		}
		xText.addText(this.elementName, elementStyle);
// add newline unless this has single PCDATANode child
		boolean addCR = true;
		if (getPCDATAContent() != null)	{
			addCR = false;
		}
		if (addCR) {
			xText.addText("\n");
		} else {
			xText.addText(": ");
		}
		Enumeration childs = children();
		while (childs.hasMoreElements()) {
			XNode child = (XNode) childs.nextElement();
			child.addContent(xText, level++);
		}
		xText.addText("\n");
	}

/** delete all whitespace descendants(uses PCDATANode.isWhitespace())
Once deleted, these cannot be returned.
(the display is not updated, since I don't
have a strong enough grasp of the MVC relationship yet. Therefore
best place to do it is after SAXTree.
<BR>
Modifies the children of this XNode. JTree may have to be recreated
if necessary.
*/
	public void deleteWhitespaceDescendants() {
		Enumeration childs = this.children();
// cannot remove elements while enumerating, so store
		Vector deleteVector = new Vector();
		while (childs.hasMoreElements()) {
			Object child = childs.nextElement();
			if (child instanceof XNode) {
				XNode xNode = (XNode) child;
				if (child instanceof PCDATANode) {
					if (((PCDATANode)child).isWhitespace()) {
						deleteVector.addElement(xNode);
					}
				} else {
					xNode.deleteWhitespaceDescendants();
				}
			}
		}
		for (int i = 0; i < deleteVector.size(); i++) {
			this.remove((XNode)deleteVector.elementAt(i));
		}
	}


/** display displayComponent in free window, probably obsolete
*/
	public void displayInFrame() {
		JFrame jFrame = new JFrame();
		jFrame.getContentPane().add(this.getDisplayComponent(CONTENT));
		jFrame.setSize(300, 300);
		jFrame.pack();
		jFrame.show();
	}

/** probably obsolete (or at least to be subclassed)
*/
	public void display(TextArea ta) {
		Object object = getUserObject();
	}
	
/** get any Font associated with the XNode (Fonts and Colors are still
being developed).
@return Font the font
*/
	public Font getFont() {return font;}
/** get any default Font associated with an XNode 
@return Font the font
*/
	public static Font getDefaultFont() {return defaultFont;}
/** set the Font associated with an XNode 
@param Font the font
*/
	public void setFont(Font font) {this.font = font;}
	
/** get any Color associated with the XNode (Fonts and Colors are still
being developed).
@return Color the color
*/
	public Color getColor() {return color;}
/** get any default Color associated with an XNode 
@return Color the color
*/
	public static Color getDefaultColor() {return defaultColor;}
/** set the Color associated with an XNode 
@param Color the color
*/
	public void setColor(Color color) {this.color = color;}

/** get an Icon associated with the XNode (still being developed).
@return Icon the icon (possibly an XMLIcon)
*/
	public Icon getIcon() {
		if (xIcon == null) {
			xIcon = new XMLIcon(this);
		}
		return xIcon;
	}

/** used by a variety of Swing routines. Simply calls getTitle().
@return String result of getTitle()
*/
	public String toString() {
		return getTitle();
	}

}

/** a class to provide XML-specific customisable icons for 
TreeCellRenderer, etc.
*/
class XMLIcon implements Icon {
	static Color backgroundColor = new Color(255, 220, 220);

	XNode xNode = null;
	String title = null;
	int height = 10;	// initial value to avoid problems
	int width = 100;

	public XMLIcon(XNode xNode) {
		this.xNode = xNode;
	}
    public void paintIcon (Component c, Graphics g, int x, int y) {
		if (title == null) {
			title = xNode.getTitle();
			title = title.substring(0, Math.min(10, title.length()));
			FontMetrics fm = g.getFontMetrics();
			height = fm.getHeight();
			width = fm.charsWidth(title.toCharArray(), 0, title.length());
		}
        g.setColor(backgroundColor);
        g.fillRect (x, y, getIconWidth(), getIconHeight());
	    g.setColor(Color.black);
		g.drawString(title, x, y+getIconHeight()-1);
    }
    public int getIconWidth() {
        return width;
    }
    public int getIconHeight() { 
        return height;
    }
}
